@@ -1,4 +1,4 @@
-# Copyright (c) 2020 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2020-2022 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -24,15 +24,30 @@ import os
 import re
 import shlex
 import socket
+import subprocess
 
 from hashlib import sha1
 from pathlib import Path, PurePath
+from subprocess import DEVNULL
+from typing import Callable, Dict, List, NoReturn, Optional, Sequence
+from typing import Set, Tuple, Union, cast
 
 from .constants import DEFAULT_PORT
-from .misc import ip_address
+from .logging import logger
+from .misc import DefTuple, FilePath, ip_address
 from .pattern import HostPatternList, WildcardPatternList
 
 
+ConfigPaths = Union[None, FilePath, Sequence[FilePath]]
+
+
+def _exec(cmd: str) -> bool:
+    """Execute a command and return if exit status is 0"""
+
+    return subprocess.run(cmd, check=False, shell=True, stdin=DEVNULL,
+                          stdout=DEVNULL, stderr=DEVNULL).returncode == 0
+
+
 class ConfigParseError(ValueError):
     """Configuration parsing exception"""
 
@@ -41,50 +56,46 @@ class SSHConfig:
     """Settings from an OpenSSH config file"""
 
     _conditionals = {'match'}
-    _no_split = set()
+    _no_split: Set[str] = set()
     _percent_expand = {'AuthorizedKeysFile'}
-    _handlers = {}
+    _handlers: Dict[str, Tuple[str, Callable]] = {}
 
-    def __init__(self, last_config, reload, *args, **kwargs):
+    def __init__(self, last_config: Optional['SSHConfig'], reload: bool):
         if last_config:
             self._last_options = last_config.get_options(reload)
         else:
             self._last_options = {}
 
-        self._path = ''
+        self._default_path = Path('~', '.ssh').expanduser()
+        self._path = Path()
         self._line_no = 0
         self._matching = True
         self._options = self._last_options.copy()
-        self._tokens = {}
+        self._tokens: Dict[str, str] = {}
 
-        self._init_options(*args, **kwargs)
+        self.loaded = False
 
-    def _init_options(self):
-        """Set options used for matching and token substitutions"""
-
-        raise NotImplementedError
-
-    def _error(self, reason, *args):
+    def _error(self, reason: str, *args: object) -> NoReturn:
         """Raise a configuration parsing error"""
 
         raise ConfigParseError('%s line %s: %s' % (self._path, self._line_no,
                                                    reason % args))
 
-    def _match_val(self, match):
+    def _match_val(self, match: str) -> object:
         """Return the value to match against in a match condition"""
 
         raise NotImplementedError
 
-    def _set_tokens(self):
+    def _set_tokens(self) -> None:
         """Set the tokens available for percent expansion"""
 
         raise NotImplementedError
 
-    def _expand_val(self, value):
+    def _expand_val(self, value: str) -> str:
         """Perform percent token expansion on a string"""
 
         last_idx = 0
-        result = []
+        result: List[str] = []
 
         for match in re.finditer(r'%', value):
             idx = match.start()
@@ -93,38 +104,52 @@ class SSHConfig:
                 continue
 
             try:
-                result.extend([value[last_idx:idx],
-                               self._tokens[value[idx+1]]])
+                token = value[idx+1]
+                result.extend([value[last_idx:idx], self._tokens[token]])
                 last_idx = idx + 2
             except IndexError:
                 raise ConfigParseError('Invalid token substitution') from None
             except KeyError:
-                raise ConfigParseError('Invalid token substitution: %s' %
-                                       value[idx+1]) from None
+                if token == 'd':
+                    raise ConfigParseError('Home directory is '
+                                           'not available') from None
+                elif token == 'i':
+                    raise ConfigParseError('User id not available') from None
+                else:
+                    raise ConfigParseError('Invalid token substitution: %s' %
+                                           value[idx+1]) from None
 
         result.append(value[last_idx:])
         return ''.join(result)
 
-    def _include(self, option, args):
+    def _include(self, option: str, args: List[str]) -> None:
         """Read config from a list of other config files"""
 
         # pylint: disable=unused-argument
 
+        old_path = self._path
+
         for pattern in args:
-            path = Path(pattern)
+            path = Path(pattern).expanduser()
 
             if path.anchor:
                 pattern = str(Path(*path.parts[1:]))
                 path = Path(path.anchor)
             else:
-                path = Path(self._path).parent
+                path = self._default_path
 
-            for path in path.glob(pattern):
+            paths = list(path.glob(pattern))
+
+            if not paths:
+                logger.debug1('Config pattern "%s" matched no files', pattern)
+
+            for path in paths:
                 self.parse(path)
 
+        self._path = old_path
         args.clear()
 
-    def _match(self, option, args):
+    def _match(self, option: str, args: List[str]) -> None:
         """Begin a conditional block"""
 
         # pylint: disable=unused-argument
@@ -138,17 +163,20 @@ class SSHConfig:
 
             match_val = self._match_val(match)
 
-            if match_val is None:
+            if match != 'exec' and match_val is None:
                 self._error('Invalid match condition')
 
             try:
-                if match in ('address', 'localaddress'):
-                    pattern = HostPatternList(args.pop(0))
-                    ip = ip_address(match_val) if match_val else None
-                    self._matching = pattern.matches(None, match_val, ip)
+                if match == 'exec':
+                    self._matching = _exec(args.pop(0))
+                elif match in ('address', 'localaddress'):
+                    host_pat = HostPatternList(args.pop(0))
+                    ip = ip_address(cast(str, match_val)) \
+                        if match_val else None
+                    self._matching = host_pat.matches(None, match_val, ip)
                 else:
-                    pattern = WildcardPatternList(args.pop(0))
-                    self._matching = pattern.matches(match_val)
+                    wild_pat = WildcardPatternList(args.pop(0))
+                    self._matching = wild_pat.matches(match_val)
             except IndexError:
                 self._error('Missing %s match pattern', match)
 
@@ -156,60 +184,62 @@ class SSHConfig:
                 args.clear()
                 break
 
-    def _set_bool(self, option, args):
+    def _set_bool(self, option: str, args: List[str]) -> None:
         """Set a boolean config option"""
 
-        value = args.pop(0).lower()
+        value_str = args.pop(0).lower()
 
-        if value in ('yes', 'true'):
+        if value_str in ('yes', 'true'):
             value = True
-        elif value in ('no', 'false'):
+        elif value_str in ('no', 'false'):
             value = False
         else:
-            self._error('Invalid %s boolean value: %s', option, value)
+            self._error('Invalid %s boolean value: %s', option, value_str)
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_int(self, option, args):
+    def _set_int(self, option: str, args: List[str]) -> None:
         """Set an integer config option"""
 
-        value = args.pop(0)
+        value_str = args.pop(0)
 
         try:
-            value = int(value)
+            value = int(value_str)
         except ValueError:
-            self._error('Invalid %s integer value: %s', option, value)
+            self._error('Invalid %s integer value: %s', option, value_str)
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_string(self, option, args):
+    def _set_string(self, option: str, args: List[str]) -> None:
         """Set a string config option"""
 
-        value = args.pop(0)
+        value_str = args.pop(0)
 
-        if value.lower() == 'none':
+        if value_str.lower() == 'none':
             value = None
+        else:
+            value = value_str
 
         if option not in self._options:
             self._options[option] = value
 
-    def _append_string(self, option, args):
+    def _append_string(self, option: str, args: List[str]) -> None:
         """Append a string config option to a list"""
 
-        value = args.pop(0)
+        value_str = args.pop(0)
 
-        if value.lower() != 'none':
+        if value_str.lower() != 'none':
             if option in self._options:
-                self._options[option].append(value)
+                cast(List[str], self._options[option]).append(value_str)
             else:
-                self._options[option] = [value]
+                self._options[option] = [value_str]
         else:
             if option not in self._options:
                 self._options[option] = []
 
-    def _set_string_list(self, option, args):
+    def _set_string_list(self, option: str, args: List[str]) -> None:
         """Set whitespace-separated string config options as a list"""
 
         if option not in self._options:
@@ -217,43 +247,43 @@ class SSHConfig:
 
         args.clear()
 
-    def _append_string_list(self, option, args):
+    def _append_string_list(self, option: str, args: List[str]) -> None:
         """Append whitespace-separated string config options to a list"""
 
         if option in self._options:
-            self._options[option].extend(args)
+            cast(List[str], self._options[option]).extend(args)
         else:
             self._options[option] = args[:]
 
         args.clear()
 
-    def _set_address_family(self, option, args):
+    def _set_address_family(self, option: str, args: List[str]) -> None:
         """Set an address family config option"""
 
-        value = args.pop(0).lower()
+        value_str = args.pop(0).lower()
 
-        if value == 'any':
+        if value_str == 'any':
             value = socket.AF_UNSPEC
-        elif value == 'inet':
+        elif value_str == 'inet':
             value = socket.AF_INET
-        elif value == 'inet6':
+        elif value_str == 'inet6':
             value = socket.AF_INET6
         else:
-            self._error('Invalid %s value: %s', option, value)
+            self._error('Invalid %s value: %s', option, value_str)
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_rekey_limits(self, option, args):
+    def _set_rekey_limits(self, option: str, args: List[str]) -> None:
         """Set rekey limits config option"""
 
-        byte_limit = args.pop(0).lower()
+        byte_limit: Union[str, Tuple[()]] = args.pop(0).lower()
 
         if byte_limit == 'default':
             byte_limit = ()
 
         if args:
-            time_limit = args.pop(0).lower()
+            time_limit: Optional[Union[str, Tuple[()]]] = args.pop(0).lower()
 
             if time_limit == 'none':
                 time_limit = None
@@ -263,7 +293,7 @@ class SSHConfig:
         if option not in self._options:
             self._options[option] = byte_limit, time_limit
 
-    def parse(self, path):
+    def parse(self, path: Path) -> None:
         """Parse an OpenSSH config file and return matching declarations"""
 
         self._path = path
@@ -271,6 +301,8 @@ class SSHConfig:
         self._matching = True
         self._tokens = {'%': '%'}
 
+        logger.debug1('Reading config from "%s"', path)
+
         with open(path) as file:
             for line in file:
                 self._line_no += 1
@@ -285,6 +317,17 @@ class SSHConfig:
                     self._error(str(exc))
 
                 option = args.pop(0)
+
+                if option.endswith('='):
+                    option = option[:-1]
+                elif '=' in option:
+                    option, arg = option.split('=', 1)
+                    args[:0] =[arg]
+                elif args and args[0] == '=':
+                    del args[0]
+                elif args and args[0].startswith('='):
+                    args[0] = args[0][1:]
+
                 loption = option.lower()
 
                 if loption in self._no_split:
@@ -316,43 +359,49 @@ class SSHConfig:
             else:
                 if isinstance(value, list):
                     value = [self._expand_val(item) for item in value]
-                else:
+                elif isinstance(value, str):
                     value = self._expand_val(value)
 
                 self._options[option] = value
 
-    def get_options(self, reload):
+    def get_options(self, reload: bool) -> Dict[str, object]:
         """Return options to base a new config object on"""
 
         return self._last_options.copy() if reload else self._options.copy()
 
     @classmethod
-    def load(cls, last_config, config_paths, reload, *args, **kwargs):
+    def load(cls, last_config: Optional['SSHConfig'],
+             config_paths: ConfigPaths, reload: bool,
+             *args: object) -> 'SSHConfig':
         """Load a list of OpenSSH config files into a config object"""
 
-        config = cls(last_config, reload, *args, **kwargs)
+        config = cls(last_config, reload, *args)
 
         if config_paths:
-            if isinstance(config_paths, (str, bytes, PurePath)):
-                config_paths = [config_paths]
+            if isinstance(config_paths, (str, PurePath)):
+                paths: Sequence[FilePath] = [config_paths]
+            else:
+                paths = config_paths
 
-            for path in config_paths:
-                config.parse(path)
+            for path in paths:
+                config.parse(Path(path))
+
+            config.loaded = True
 
         return config
 
-    def get(self, option, default=None):
+    def get(self, option: str, default: object = None) -> object:
         """Get the value of a config option"""
 
         return self._options.get(option, default)
 
-    def get_compression_algs(self, default=None):
+    def get_compression_algs(self) -> DefTuple[str]:
         """Return the compression algorithms to use"""
 
         compression = self.get('Compression')
 
         if compression is None:
-            return default
+            return ()
         elif compression:
             return 'zlib@openssh.com,zlib,none'
         else:
@@ -365,12 +414,11 @@ class SSHClientConfig(SSHConfig):
     _conditionals = {'host', 'match'}
     _no_split = {'remotecommand'}
     _percent_expand = {'CertificateFile', 'IdentityAgent',
-                       'IdentityFile', 'RemoteCommand'}
+                       'IdentityFile', 'ProxyCommand', 'RemoteCommand'}
 
-    def _init_options(self, local_user, user, host, port):
-        """Set options used for matching and token substitutions"""
-
-        # pylint: disable=arguments-differ
+    def __init__(self, last_config: 'SSHConfig', reload: bool,
+                 local_user: str, user: str, host: str, port: int) -> None:
+        super().__init__(last_config, reload)
 
         self._local_user = local_user
         self._orig_host = host
@@ -381,7 +429,7 @@ class SSHClientConfig(SSHConfig):
         if port != ():
             self._options['Port'] = port
 
-    def _match_val(self, match):
+    def _match_val(self, match: str) -> object:
         """Return the value to match against in a match condition"""
 
         if match == 'host':
@@ -395,7 +443,7 @@ class SSHClientConfig(SSHConfig):
         else:
             return None
 
-    def _match_host(self, option, args):
+    def _match_host(self, option: str, args: List[str]) -> None:
         """Begin a conditional block matching on host"""
 
         # pylint: disable=unused-argument
@@ -404,31 +452,34 @@ class SSHClientConfig(SSHConfig):
         self._matching = WildcardPatternList(pattern).matches(self._orig_host)
         args.clear()
 
-    def _set_hostname(self, option, args):
+    def _set_hostname(self, option: str, args: List[str]) -> None:
         """Set hostname config option"""
 
         value = args.pop(0)
 
         if option not in self._options:
-            self._tokens['h'] = self._options.get(option, self._orig_host)
+            self._tokens['h'] = \
+                cast(str, self._options.get(option, self._orig_host))
             self._options[option] = self._expand_val(value)
 
-    def _set_request_tty(self, option, args):
+    def _set_request_tty(self, option: str, args: List[str]) -> None:
         """Set a pseudo-terminal request config option"""
 
-        value = args.pop(0).lower()
+        value_str = args.pop(0).lower()
 
-        if value in ('yes', 'true'):
-            value = True
-        elif value in ('no', 'false'):
+        if value_str in ('yes', 'true'):
+            value: Union[bool, str] = True
+        elif value_str in ('no', 'false'):
             value = False
-        elif value not in ('force', 'auto'):
-            self._error('Invalid %s value: %s', option, value)
+        elif value_str not in ('force', 'auto'):
+            self._error('Invalid %s value: %s', option, value_str)
+        else:
+            value = value_str
 
         if option not in self._options:
             self._options[option] = value
 
-    def _set_tokens(self):
+    def _set_tokens(self) -> None:
         """Set the tokens available for percent expansion"""
 
         local_host = socket.gethostname()
@@ -436,15 +487,15 @@ class SSHClientConfig(SSHConfig):
         idx = local_host.find('.')
         short_local_host = local_host if idx < 0 else local_host[:idx]
 
-        host = self._options.get('Hostname', self._orig_host)
+        host = cast(str, self._options.get('Hostname', self._orig_host))
         port = str(self._options.get('Port', DEFAULT_PORT))
-        user = self._options.get('User') or self._local_user
+        user = cast(str, self._options.get('User') or self._local_user)
+        home = os.path.expanduser('~')
 
         conn_info = ''.join((local_host, host, port, user))
         conn_hash = sha1(conn_info.encode('utf-8')).hexdigest()
 
         self._tokens.update({'C': conn_hash,
-                             'd': str(Path.home()),
                              'h': host,
                              'L': short_local_host,
                              'l': local_host,
@@ -453,10 +504,11 @@ class SSHClientConfig(SSHConfig):
                              'r': user,
                              'u': self._local_user})
 
-        if hasattr(os, 'getuid'): # pragma: no branch
-            self._tokens['i'] = str(os.getuid())
+        if home != '~':
+            self._tokens['d'] = home
 
-    # pylint: disable=bad-whitespace
+        if hasattr(os, 'getuid'):
+            self._tokens['i'] = str(os.getuid())
 
     _handlers = {option.lower(): (option, handler) for option, handler in (
         ('Host',                            _match_host),
@@ -481,6 +533,7 @@ class SSHClientConfig(SSHConfig):
         ('HostbasedAuthentication',         SSHConfig._set_bool),
         ('HostKeyAlgorithms',               SSHConfig._set_string),
         ('Hostname',                        _set_hostname),
+        ('HostKeyAlias',                    SSHConfig._set_string),
         ('IdentitiesOnly',                  SSHConfig._set_bool),
         ('IdentityAgent',                   SSHConfig._set_string),
         ('IdentityFile',                    SSHConfig._append_string),
@@ -491,6 +544,7 @@ class SSHClientConfig(SSHConfig):
         ('PKCS11Provider',                  SSHConfig._set_string),
         ('PreferredAuthentications',        SSHConfig._set_string),
         ('Port',                            SSHConfig._set_int),
+        ('ProxyCommand',                    SSHConfig._set_string_list),
         ('ProxyJump',                       SSHConfig._set_string),
         ('PubkeyAuthentication',            SSHConfig._set_bool),
         ('RekeyLimit',                      SSHConfig._set_rekey_limits),
@@ -509,10 +563,10 @@ class SSHClientConfig(SSHConfig):
 class SSHServerConfig(SSHConfig):
     """Settings from an OpenSSH server config file"""
 
-    def _init_options(self, local_addr, local_port, user, host, addr):
-        """Set options used for matching and token substitutions"""
-
-        # pylint: disable=arguments-differ
+    def __init__(self, last_config: 'SSHConfig', reload: bool,
+                 local_addr: str, local_port: int, user: str,
+                 host: str, addr: str) -> None:
+        super().__init__(last_config, reload)
 
         self._local_addr = local_addr
         self._local_port = local_port
@@ -520,7 +574,7 @@ class SSHServerConfig(SSHConfig):
         self._host = host or addr
         self._addr = addr
 
-    def _match_val(self, match):
+    def _match_val(self, match: str) -> object:
         """Return the value to match against in a match condition"""
 
         if match == 'localaddress':
@@ -536,13 +590,11 @@ class SSHServerConfig(SSHConfig):
         else:
             return None
 
-    def _set_tokens(self):
+    def _set_tokens(self) -> None:
         """Set the tokens available for percent expansion"""
 
         self._tokens.update({'u': self._user})
 
-    # pylint: disable=bad-whitespace
-
     _handlers = {option.lower(): (option, handler) for option, handler in (
         ('Match',                           SSHConfig._match),
         ('Include',                         SSHConfig._include),
