@@ -1,22 +1,51 @@
-# Copyright (c) 2013-2015 by Ron Frederick <ronf@timeheart.net>.
-# All rights reserved.
+# Copyright (c) 2013-2021 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
-# the terms of the Eclipse Public License v1.0 which accompanies this
+# the terms of the Eclipse Public License v2.0 which accompanies this
 # distribution and is available at:
 #
-#     http://www.eclipse.org/legal/epl-v10.html
+#     http://www.eclipse.org/legal/epl-2.0/
+#
+# This program may also be made available under the following secondary
+# licenses when the conditions for such availability set forth in the
+# Eclipse Public License v2.0 are satisfied:
+#
+#    GNU General Public License, Version 2.0, or any later versions of
+#    that license
+#
+# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
 #
 # Contributors:
 #     Ron Frederick - initial implementation, API, and documentation
 
 """SSH session handlers"""
 
+from typing import TYPE_CHECKING, Any, AnyStr, Callable, Generic
+from typing import Mapping, Optional, Tuple, Union
 
-class SSHSession:
+
+if TYPE_CHECKING:
+    # pylint: disable=cyclic-import
+    from .channel import SSHClientChannel, SSHServerChannel
+    from .channel import SSHTCPChannel, SSHUNIXChannel
+
+DataType = Optional[int]
+
+TermModes = Mapping[int, int]
+TermModesArg = Optional[TermModes]
+TermSize = Tuple[int, int, int, int]
+TermSizeArg = Union[None, Tuple[int, int], TermSize]
+
+
+class SSHSession(Generic[AnyStr]):
     """SSH session handler"""
 
-    def connection_lost(self, exc):
+    # pylint: disable=no-self-use,unused-argument
+
+    def connection_made(self, chan: Any) -> None:
+        """Called when a channel is opened successfully"""
+
+    def connection_lost(self, exc: Optional[Exception]) -> None:
         """Called when a channel is closed
 
            This method is called when a channel is closed. If the channel
@@ -30,9 +59,7 @@ class SSHSession:
 
         """
 
-        pass # pragma: no cover
-
-    def session_started(self):
+    def session_started(self) -> None:
         """Called when the session is started
 
            This method is called when a session has started up. For
@@ -43,9 +70,7 @@ class SSHSession:
 
         """
 
-        pass # pragma: no cover
-
-    def data_received(self, data, datatype):
+    def data_received(self, data: AnyStr, datatype: DataType) -> None:
         """Called when data is received on the channel
 
            This method is called when data is received on the channel.
@@ -63,9 +88,7 @@ class SSHSession:
 
         """
 
-        pass # pragma: no cover
-
-    def eof_received(self):
+    def eof_received(self) -> bool:
         """Called when EOF is received on the channel
 
            This method is called when an end-of-file indication is received
@@ -81,11 +104,9 @@ class SSHSession:
 
         """
 
-        # pylint: disable=no-self-use
-
         return False # pragma: no cover
 
-    def pause_writing(self):
+    def pause_writing(self) -> None:
         """Called when the write buffer becomes full
 
            This method is called when the channel's write buffer becomes
@@ -96,9 +117,7 @@ class SSHSession:
 
         """
 
-        pass # pragma: no cover
-
-    def resume_writing(self):
+    def resume_writing(self) -> None:
         """Called when the write buffer has sufficiently drained
 
            This method is called when the channel's send window reopens
@@ -107,10 +126,8 @@ class SSHSession:
 
         """
 
-        pass # pragma: no cover
 
-
-class SSHClientSession(SSHSession):
+class SSHClientSession(SSHSession[AnyStr]):
     """SSH client session handler
 
        Applications should subclass this when implementing an SSH client
@@ -128,7 +145,9 @@ class SSHClientSession(SSHSession):
 
     """
 
-    def connection_made(self, chan):
+    # pylint: disable=no-self-use,unused-argument
+
+    def connection_made(self, chan: 'SSHClientChannel[AnyStr]') -> None:
         """Called when a channel is opened successfully
 
            This method is called when a channel is opened successfully. The
@@ -140,9 +159,7 @@ class SSHClientSession(SSHSession):
 
         """
 
-        pass # pragma: no cover
-
-    def xon_xoff_requested(self, client_can_do):
+    def xon_xoff_requested(self, client_can_do: bool) -> None:
         """XON/XOFF flow control has been enabled or disabled
 
            This method is called to notify the client whether or not
@@ -163,9 +180,7 @@ class SSHClientSession(SSHSession):
 
         """
 
-        pass # pragma: no cover
-
-    def exit_status_received(self, status):
+    def exit_status_received(self, status: int) -> None:
         """A remote exit status has been received for this session
 
            This method is called when the shell, command, or subsystem
@@ -182,9 +197,8 @@ class SSHClientSession(SSHSession):
 
         """
 
-        pass # pragma: no cover
-
-    def exit_signal_received(self, signal, core_dumped, msg, lang):
+    def exit_signal_received(self, signal: str, core_dumped: bool,
+                             msg: str, lang: str) -> None:
         """A remote exit signal has been received for this session
 
            This method is called when the shell, command, or subsystem
@@ -210,10 +224,8 @@ class SSHClientSession(SSHSession):
 
         """
 
-        pass # pragma: no cover
 
-
-class SSHServerSession(SSHSession):
+class SSHServerSession(SSHSession[AnyStr]):
     """SSH server session handler
 
        Applications should subclass this when implementing an SSH server
@@ -233,7 +245,9 @@ class SSHServerSession(SSHSession):
 
     """
 
-    def connection_made(self, chan):
+    # pylint: disable=no-self-use,unused-argument
+
+    def connection_made(self, chan: 'SSHServerChannel[AnyStr]') -> None:
         """Called when a channel is opened successfully
 
            This method is called when a channel is opened successfully. The
@@ -245,9 +259,9 @@ class SSHServerSession(SSHSession):
 
         """
 
-        pass # pragma: no cover
-
-    def pty_requested(self, term_type, term_size, term_modes):
+    def pty_requested(self, term_type: str,
+                      term_size: Tuple[int, int, int, int],
+                      term_modes: Mapping[int, int]) -> bool:
         """A psuedo-terminal has been requested
 
            This method is called when the client sends a request to allocate
@@ -286,11 +300,10 @@ class SSHServerSession(SSHSession):
 
         """
 
-        # pylint: disable=no-self-use,unused-argument
-
         return True # pragma: no cover
 
-    def terminal_size_changed(self, width, height, pixwidth, pixheight):
+    def terminal_size_changed(self, width: int, height: int,
+                              pixwidth: int, pixheight: int) -> None:
         """The terminal size has changed
 
            This method is called when a client requests a
@@ -316,11 +329,7 @@ class SSHServerSession(SSHSession):
 
         """
 
-        # pylint: disable=no-self-use,unused-argument
-
-        pass # pragma: no cover
-
-    def shell_requested(self):
+    def shell_requested(self) -> bool:
         """The client has requested a shell
 
            This method should be implemented by the application to
@@ -339,11 +348,9 @@ class SSHServerSession(SSHSession):
 
         """
 
-        # pylint: disable=no-self-use,unused-argument
-
         return False # pragma: no cover
 
-    def exec_requested(self, command):
+    def exec_requested(self, command: str) -> bool:
         """The client has requested to execute a command
 
            This method should be implemented by the application to
@@ -366,11 +373,9 @@ class SSHServerSession(SSHSession):
 
         """
 
-        # pylint: disable=no-self-use,unused-argument
-
         return False # pragma: no cover
 
-    def subsystem_requested(self, subsystem):
+    def subsystem_requested(self, subsystem: str) -> bool:
         """The client has requested to start a subsystem
 
            This method should be implemented by the application to
@@ -393,11 +398,9 @@ class SSHServerSession(SSHSession):
 
         """
 
-        # pylint: disable=no-self-use,unused-argument
-
         return False # pragma: no cover
 
-    def break_received(self, msec):
+    def break_received(self, msec: int) -> bool:
         """The client has sent a break
 
            This method is called when the client requests that the
@@ -417,11 +420,9 @@ class SSHServerSession(SSHSession):
 
         """
 
-        # pylint: disable=no-self-use,unused-argument
-
         return False # pragma: no cover
 
-    def signal_received(self, signal):
+    def signal_received(self, signal: str) -> None:
         """The client has sent a signal
 
            This method is called when the client delivers a signal
@@ -429,14 +430,26 @@ class SSHServerSession(SSHSession):
 
            By default, signals from the client are ignored.
 
+           :param signal:
+               The name of the signal received
+           :type signal: `str`
+
         """
 
-        # pylint: disable=no-self-use,unused-argument
+    def soft_eof_received(self) -> None:
+        """The client has sent a soft EOF
 
-        pass # pragma: no cover
+           This method is called by the line editor when the client
+           send a soft EOF (Ctrl-D on an empty input line).
+
+           By default, soft EOF will trigger an EOF to an outstanding
+           read call but still allow additional input to be received
+           from the client after that.
+
+        """
 
 
-class SSHTCPSession(SSHSession):
+class SSHTCPSession(SSHSession[AnyStr]):
     """SSH TCP session handler
 
        Applications should subclass this when implementing a handler for
@@ -466,7 +479,7 @@ class SSHTCPSession(SSHSession):
 
     """
 
-    def connection_made(self, chan):
+    def connection_made(self, chan: 'SSHTCPChannel[AnyStr]') -> None:
         """Called when a channel is opened successfully
 
            This method is called when a channel is opened successfully. The
@@ -478,10 +491,8 @@ class SSHTCPSession(SSHSession):
 
         """
 
-        pass # pragma: no cover
 
-
-class SSHUNIXSession(SSHSession):
+class SSHUNIXSession(SSHSession[AnyStr]):
     """SSH UNIX domain socket session handler
 
        Applications should subclass this when implementing a handler for
@@ -513,7 +524,7 @@ class SSHUNIXSession(SSHSession):
 
     """
 
-    def connection_made(self, chan):
+    def connection_made(self, chan: 'SSHUNIXChannel[AnyStr]') -> None:
         """Called when a channel is opened successfully
 
            This method is called when a channel is opened successfully. The
@@ -525,4 +536,8 @@ class SSHUNIXSession(SSHSession):
 
         """
 
-        pass # pragma: no cover
+
+SSHSessionFactory = Callable[[], SSHSession[AnyStr]]
+SSHClientSessionFactory = Callable[[], SSHClientSession[AnyStr]]
+SSHTCPSessionFactory = Callable[[], SSHTCPSession[AnyStr]]
+SSHUNIXSessionFactory = Callable[[], SSHUNIXSession[AnyStr]]
