@@ -1,4 +1,4 @@
-# Copyright (c) 2019-2020 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2019-2021 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -21,25 +21,33 @@
 """U2F EdDSA public key encryption handler"""
 
 from hashlib import sha256
+from typing import Optional, Tuple, cast
 
 from .crypto import EdDSAPublicKey, ed25519_available
-from .packet import Byte, String, UInt32
+from .packet import Byte, String, UInt32, SSHPacket
 from .public_key import KeyExportError, SSHKey, SSHOpenSSHCertificateV01
 from .public_key import register_public_key_alg, register_certificate_alg
 from .public_key import register_sk_alg
 from .sk import SSH_SK_ED25519, SSH_SK_USER_PRESENCE_REQD, sk_enroll, sk_sign
 
 
+_PrivateKeyArgs = Tuple[bytes, bytes, int, bytes, bytes]
+_PublicKeyArgs = Tuple[bytes, bytes]
+
+
 class _SKEd25519Key(SSHKey):
     """Handler for U2F Ed25519 public key encryption"""
 
+    _key: EdDSAPublicKey
+
     algorithm = b'sk-ssh-ed25519@openssh.com'
     sig_algorithms = (algorithm,)
     all_sig_algorithms = set(sig_algorithms)
     use_executor = True
 
-    def __init__(self, public_value, application, flags=None,
-                 key_handle=None, reserved=None):
+    def __init__(self, public_value: bytes, application: bytes,
+                 flags: int = 0, key_handle: bytes = None,
+                 reserved: bytes = b''):
         super().__init__(EdDSAPublicKey.construct(b'ed25519', public_value))
 
         self._application = application
@@ -47,7 +55,7 @@ class _SKEd25519Key(SSHKey):
         self._key_handle = key_handle
         self._reserved = reserved
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         # This isn't protected access - both objects are _SKEd25519Key instances
         # pylint: disable=protected-access
 
@@ -58,16 +66,18 @@ class _SKEd25519Key(SSHKey):
                 self._key_handle == other._key_handle and
                 self._reserved == other._reserved)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash((self._key.public_value, self._application, self._flags,
                      self._key_handle, self._reserved))
 
     @classmethod
-    def generate(cls, algorithm, application='ssh:', user='AsyncSSH',
-                 pin=None, resident=False, touch_required=True):
+    def generate(cls, algorithm: bytes, *, # type: ignore
+                 application: str = 'ssh:', user: str = 'AsyncSSH',
+                 pin: Optional[str] = None, resident: bool = False,
+                 touch_required: bool = True) -> '_SKEd25519Key':
         """Generate a new U2F Ed25519 private key"""
 
-        # pylint: disable=unused-argument
+        # pylint: disable=arguments-differ
 
         application = application.encode('utf-8')
         flags = SSH_SK_USER_PRESENCE_REQD if touch_required else 0
@@ -78,20 +88,24 @@ class _SKEd25519Key(SSHKey):
         return cls(public_value, application, flags, key_handle, b'')
 
     @classmethod
-    def make_private(cls, public_value, application,
-                     flags, key_handle, reserved):
+    def make_private(cls, key_params: object) -> SSHKey:
         """Construct a U2F Ed25519 private key"""
 
+        public_value, application, flags, key_handle, reserved = \
+            cast(_PrivateKeyArgs, key_params)
+
         return cls(public_value, application, flags, key_handle, reserved)
 
     @classmethod
-    def make_public(cls, public_value, application):
+    def make_public(cls, key_params: object) -> SSHKey:
         """Construct a U2F Ed25519 public key"""
 
+        public_value, application = cast(_PublicKeyArgs, key_params)
+
         return cls(public_value, application)
 
     @classmethod
-    def decode_ssh_private(cls, packet):
+    def decode_ssh_private(cls, packet: SSHPacket) -> _PrivateKeyArgs:
         """Decode an SSH format U2F Ed25519 private key"""
 
         public_value = packet.get_string()
@@ -103,7 +117,7 @@ class _SKEd25519Key(SSHKey):
         return public_value, application, flags, key_handle, reserved
 
     @classmethod
-    def decode_ssh_public(cls, packet):
+    def decode_ssh_public(cls, packet: SSHPacket) -> _PublicKeyArgs:
         """Decode an SSH format U2F Ed25519 public key"""
 
         public_value = packet.get_string()
@@ -111,7 +125,7 @@ class _SKEd25519Key(SSHKey):
 
         return public_value, application
 
-    def encode_ssh_private(self):
+    def encode_ssh_private(self) -> bytes:
         """Encode an SSH format U2F Ed25519 private key"""
 
         if self._key_handle is None:
@@ -121,18 +135,18 @@ class _SKEd25519Key(SSHKey):
                          String(self._application), Byte(self._flags),
                          String(self._key_handle), String(self._reserved)))
 
-    def encode_ssh_public(self):
+    def encode_ssh_public(self) -> bytes:
         """Encode an SSH format U2F Ed25519 public key"""
 
         return b''.join((String(self._key.public_value),
                          String(self._application)))
 
-    def encode_agent_cert_private(self):
+    def encode_agent_cert_private(self) -> bytes:
         """Encode U2F Ed25519 certificate private key data for agent"""
 
         return self.encode_ssh_private()
 
-    def sign_ssh(self, data, sig_algorithm):
+    def sign_ssh(self, data: bytes, sig_algorithm: bytes) -> bytes:
         """Compute an SSH-encoded signature of the specified data"""
 
         # pylint: disable=unused-argument
@@ -145,7 +159,8 @@ class _SKEd25519Key(SSHKey):
 
         return String(sig) + Byte(flags) + UInt32(counter)
 
-    def verify_ssh(self, data, sig_algorithm, packet):
+    def verify_ssh(self, data: bytes, sig_algorithm: bytes,
+                   packet: SSHPacket) -> bool:
         """Verify an SSH-encoded signature of the specified data"""
 
         # pylint: disable=unused-argument
