@@ -1,6 +1,6 @@
 #!/usr/bin/env python3.6
 #
-# Copyright (c) 2013-2018 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2013-2021 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -22,7 +22,7 @@
 
 import asyncio, asyncssh, sys
 
-async def run_client():
+async def run_client() -> None:
     async with asyncssh.connect('localhost') as conn:
         result = await conn.run('ls abc')
 
@@ -30,7 +30,7 @@ async def run_client():
             print(result.stdout, end='')
         else:
             print(result.stderr, end='', file=sys.stderr)
-            print('Program exited with status %d' % result.exit_status,
+            print(f'Program exited with status {result.exit_status}',
                   file=sys.stderr)
 
 try:
