@@ -1,13 +1,21 @@
-#!/usr/bin/env python3.5
+#!/usr/bin/env python3.6
 #
-# Copyright (c) 2016 by Ron Frederick <ronf@timeheart.net>.
-# All rights reserved.
+# Copyright (c) 2016-2021 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
-# the terms of the Eclipse Public License v1.0 which accompanies this
+# the terms of the Eclipse Public License v2.0 which accompanies this
 # distribution and is available at:
 #
-#     http://www.eclipse.org/legal/epl-v10.html
+#     http://www.eclipse.org/legal/epl-2.0/
+#
+# This program may also be made available under the following secondary
+# licenses when the conditions for such availability set forth in the
+# Eclipse Public License v2.0 are satisfied:
+#
+#    GNU General Public License, Version 2.0, or any later versions of
+#    that license
+#
+# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
 #
 # Contributors:
 #     Ron Frederick - initial implementation, API, and documentation
@@ -22,12 +30,12 @@
 import asyncio, asyncssh, os, sys
 
 class MySFTPServer(asyncssh.SFTPServer):
-    def __init__(self, conn):
-        root = '/tmp/sftp/' + conn.get_extra_info('username')
+    def __init__(self, chan: asyncssh.SSHServerChannel):
+        root = '/tmp/sftp/' + chan.get_extra_info('username')
         os.makedirs(root, exist_ok=True)
-        super().__init__(conn, chroot=root)
+        super().__init__(chan, chroot=root)
 
-async def start_server():
+async def start_server() -> None:
     await asyncssh.listen('', 8022, server_host_keys=['ssh_host_key'],
                           authorized_client_keys='ssh_user_ca',
                           sftp_factory=MySFTPServer)
