@@ -1,6 +1,6 @@
 #!/usr/bin/env python3.6
 #
-# Copyright (c) 2013-2018 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2013-2021 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -30,16 +30,16 @@
 
 import asyncio, asyncssh, sys
 
-def handle_client(process):
+def handle_client(process: asyncssh.SSHServerProcess) -> None:
     process.stdout.write('Welcome to my SSH server, %s!\n' %
                          process.get_extra_info('username'))
     process.exit(0)
 
 class MySSHServer(asyncssh.SSHServer):
-    def connection_made(self, conn):
+    def connection_made(self, conn: asyncssh.SSHServerConnection) -> None:
         self._conn = conn
 
-    def begin_auth(self, username):
+    def begin_auth(self, username: str) -> bool:
         try:
             self._conn.set_authorized_keys('authorized_keys/%s' % username)
         except IOError:
@@ -47,7 +47,7 @@ class MySSHServer(asyncssh.SSHServer):
 
         return True
 
-async def start_server():
+async def start_server() -> None:
     await asyncssh.create_server(MySSHServer, '', 8022,
                                  server_host_keys=['ssh_host_key'],
                                  process_factory=handle_client)
