@@ -1,4 +1,4 @@
-# Copyright (c) 2016-2020 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2016-2022 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -63,6 +63,15 @@ class _AsyncGSSServer(asyncssh.SSHServer):
                                               host_principal)
 
 
+class _NullServer(Server):
+    """Server for testing disabled auth"""
+
+    async def begin_auth(self, username):
+        """Handle client authentication request"""
+
+        return False
+
+
 class _HostBasedServer(Server):
     """Server for testing host-based authentication"""
 
@@ -361,7 +370,9 @@ class _KbdintServer(Server):
         """Return a password challenge after the instructions"""
 
         if self._kbdint_round == 0:
-            if username == 'pw':
+            if username == 'none':
+                result = ('', '', '', [])
+            elif username == 'pw':
                 result = ('', '', '', [('Password:', False)])
             elif username == 'pc':
                 result = ('', '', '', [('Passcode:', False)])
@@ -370,7 +381,7 @@ class _KbdintServer(Server):
             else:
                 result = ('', '', '', [('Other Challenge:', False)])
         else:
-            if responses in (['kbdint'], ['1', '2']):
+            if responses in ([], ['kbdint'], ['1', '2']):
                 result = True
             else:
                 result = ('', '', '', [('Second Challenge:', True)])
@@ -406,6 +417,39 @@ class _UnknownAuthClientConnection(asyncssh.connection.SSHClientConnection):
         super().try_next_auth()
 
 
+class _TestNullAuth(ServerTestCase):
+    """Unit tests for testing disabled authentication"""
+
+    @classmethod
+    async def start_server(cls):
+        """Start an SSH server which supports disabled authentication"""
+
+        return await cls.create_server(_NullServer)
+
+    @asynctest
+    async def test_get_server_auth_methods(self):
+        """Test getting auth methods from the test server"""
+
+        auth_methods = await asyncssh.get_server_auth_methods(
+            self._server_addr, self._server_port)
+
+        self.assertEqual(auth_methods, ['none'])
+
+    @asynctest
+    async def test_disabled_auth(self):
+        """Test disabled authentication"""
+
+        async with self.connect(username='user'):
+            pass
+
+    @asynctest
+    async def test_disabled_trivial_auth(self):
+        """Test disabling trivial auth with no authentication"""
+
+        with self.assertRaises(asyncssh.PermissionDenied):
+            await self.connect(username='user', disable_trivial_auth=True)
+
+
 @unittest.skipUnless(gss_available, 'GSS not available')
 @patch_gss
 class _TestGSSAuth(ServerTestCase):
@@ -417,6 +461,15 @@ class _TestGSSAuth(ServerTestCase):
 
         return await cls.create_server(_AsyncGSSServer, gss_host='1')
 
+    @asynctest
+    async def test_get_server_auth_methods(self):
+        """Test getting auth methods from the test server"""
+
+        auth_methods = await asyncssh.get_server_auth_methods(
+            self._server_addr, self._server_port)
+
+        self.assertEqual(auth_methods, ['gssapi-with-mic'])
+
     @asynctest
     async def test_gss_kex_auth(self):
         """Test GSS key exchange authentication"""
@@ -470,6 +523,24 @@ class _TestGSSAuth(ServerTestCase):
         with self.assertRaises(asyncssh.PermissionDenied):
             await self.connect(gss_host='1,init_error', username='user')
 
+    @asynctest
+    async def test_disabled_trivial_gss_kex_auth(self):
+        """Test disabling trivial auth with GSS key exchange authentication"""
+
+        async with self.connect(kex_algs=['gss-gex-sha256'],
+                                username='user', gss_host='1',
+                                disable_trivial_auth=True):
+            pass
+
+    @asynctest
+    async def test_disabled_trivial_gss_mic_auth(self):
+        """Test disabling trivial auth with GSS MIC authentication"""
+
+        async with self.connect(kex_algs=['ecdh-sha2-nistp256'],
+                                username='user', gss_host='1',
+                                disable_trivial_auth=True):
+            pass
+
 
 @unittest.skipUnless(gss_available, 'GSS not available')
 @patch_gss
@@ -562,6 +633,15 @@ class _TestHostBasedAuth(ServerTestCase):
         return await cls.create_server(
             _HostBasedServer, known_client_hosts='known_hosts')
 
+    @asynctest
+    async def test_get_server_auth_methods(self):
+        """Test getting auth methods from the test server"""
+
+        auth_methods = await asyncssh.get_server_auth_methods(
+            self._server_addr, self._server_port, username='user')
+
+        self.assertEqual(auth_methods, ['hostbased'])
+
     @asynctest
     async def test_client_host_auth(self):
         """Test connecting with host-based authentication"""
@@ -725,6 +805,15 @@ class _TestHostBasedAuth(ServerTestCase):
             await self.connect(username='user', client_host_keys=[(ckey, cert)],
                                client_username='user')
 
+    @asynctest
+    async def test_disabled_trivial_client_host_auth(self):
+        """Test disabling trivial auth with host-based authentication"""
+
+        with self.assertRaises(asyncssh.PermissionDenied):
+            await self.connect(username='user', client_host_keys='skey',
+                               client_username='user',
+                               disable_trivial_auth=True)
+
 
 @patch_getnameinfo
 class _TestCallbackHostBasedAuth(ServerTestCase):
@@ -948,6 +1037,38 @@ class _TestPublicKeyAuth(ServerTestCase):
 
         return conn
 
+    @asynctest
+    async def test_get_server_auth_methods(self):
+        """Test getting auth methods from the test server"""
+
+        auth_methods = await asyncssh.get_server_auth_methods(
+            self._server_addr, self._server_port)
+
+        self.assertEqual(auth_methods, ['publickey'])
+
+    @asynctest
+    async def test_encrypted_client_key(self):
+        """Test public key auth with encrypted client key"""
+
+        async with self.connect(username='ckey', client_keys='ckey_encrypted',
+                                passphrase='passphrase'):
+            pass
+
+    @asynctest
+    async def test_encrypted_client_key_bad_passphrase(self):
+        """Test wrong passphrase for encrypted client key"""
+
+        with self.assertRaises(asyncssh.KeyEncryptionError):
+            await self.connect(username='ckey', client_keys='ckey_encrypted',
+                               passphrase='xxx')
+
+    @asynctest
+    async def test_encrypted_client_key_missing_passphrase(self):
+        """Test missing passphrase for encrypted client key"""
+
+        with self.assertRaises(asyncssh.KeyImportError):
+            await self.connect(username='ckey', client_keys='ckey_encrypted')
+
     @asynctest
     async def test_client_certs(self):
         """Test trusted client certificate via client_certs"""
@@ -1154,6 +1275,30 @@ class _TestPublicKeyAuth(ServerTestCase):
                 async with self.connect(username='ckey', client_keys=[key]):
                     pass
 
+    @asynctest
+    async def test_keypair_with_replaced_cert(self):
+        """Test connecting with a keypair with replaced cert"""
+
+        ckey = asyncssh.load_keypairs(['ckey'])[0]
+
+        async with self.connect(username='ckey',
+                                client_keys=[(ckey, 'ckey-cert.pub')]):
+            pass
+
+    @asynctest
+    async def test_agent_keypair_with_replaced_cert(self):
+        """Test connecting with sn agent key with replaced cert"""
+
+        if not self.agent_available(): # pragma: no cover
+            self.skipTest('ssh-agent not available')
+
+        async with asyncssh.connect_agent() as agent:
+            ckey = (await agent.get_keys())[2]
+
+            async with self.connect(username='ckey',
+                                    client_keys=[(ckey, 'ckey-cert.pub')]):
+                pass
+
     @asynctest
     async def test_untrusted_client_key(self):
         """Test untrusted client key"""
@@ -1277,6 +1422,14 @@ class _TestPublicKeyAuth(ServerTestCase):
                                     agent_path=None):
                 pass
 
+    @asynctest
+    async def test_disabled_trivial_public_key_auth(self):
+        """Test disabling trivial auth with public key authentication"""
+
+        async with self.connect(username='ckey', agent_path=None,
+                                disable_trivial_auth=True):
+            pass
+
 
 class _TestPublicKeyAsyncServerAuth(_TestPublicKeyAuth):
     """Unit tests for public key authentication with async server callbacks"""
@@ -1429,6 +1582,30 @@ class _TestX509Auth(ServerTestCase):
                                 client_keys=['ckey_x509_chain']):
             pass
 
+    @asynctest
+    async def test_keypair_with_x509_cert(self):
+        """Test connecting with a keypair with replaced X.509 cert"""
+
+        ckey = asyncssh.load_keypairs(['ckey'])[0]
+
+        async with self.connect(username='ckey',
+                                client_keys=[(ckey, 'ckey_x509_chain')]):
+            pass
+
+    @asynctest
+    async def test_agent_keypair_with_x509_cert(self):
+        """Test connecting with sn agent key with replaced X.509 cert"""
+
+        if not self.agent_available(): # pragma: no cover
+            self.skipTest('ssh-agent not available')
+
+        async with asyncssh.connect_agent() as agent:
+            ckey = (await agent.get_keys())[2]
+
+            async with self.connect(username='ckey',
+                                    client_keys=[(ckey, 'ckey_x509_chain')]):
+                pass
+
     @asynctest
     async def test_x509_incomplete_chain(self):
         """Test connecting with incomplete X.509 certificate chain"""
@@ -1445,6 +1622,16 @@ class _TestX509Auth(ServerTestCase):
         with self.assertRaises(asyncssh.PermissionDenied):
             await self.connect(username='ckey', client_keys=['skey_x509_chain'])
 
+    @asynctest
+    async def test_disabled_trivial_x509_auth(self):
+        """Test disabling trivial auth with X.509 certificate authentication"""
+
+        async with self.connect(username='ckey',
+                                client_keys=['ckey_x509_self'],
+                                disable_trivial_auth=True):
+            pass
+
+
 @unittest.skipUnless(x509_available, 'X.509 not available')
 class _TestX509AuthDisabled(ServerTestCase):
     """Unit tests for disabled X.509 certificate authentication"""
@@ -1543,9 +1730,19 @@ class _TestPasswordAuth(ServerTestCase):
 
         return await cls.create_server(_PasswordServer)
 
+    @asynctest
+    async def test_get_server_auth_methods(self):
+        """Test getting auth methods from the test server"""
+
+        auth_methods = await asyncssh.get_server_auth_methods(
+            self._server_addr, self._server_port, username='pw')
+
+        self.assertEqual(auth_methods, ['keyboard-interactive', 'password'])
+
     @async_context_manager
     async def _connect_password(self, username, password, old_password='',
-                                new_password='', test_async=False):
+                                new_password='', disable_trivial_auth=False,
+                                test_async=False):
         """Open a connection to test password authentication"""
 
         def client_factory():
@@ -1554,9 +1751,9 @@ class _TestPasswordAuth(ServerTestCase):
             cls = _AsyncPasswordClient if test_async else _PasswordClient
             return cls(password, old_password, new_password)
 
-        conn, _ = await self.create_connection(client_factory,
-                                               username=username,
-                                               client_keys=None)
+        conn, _ = await self.create_connection(
+            client_factory, username=username, client_keys=None,
+            disable_trivial_auth=disable_trivial_auth)
 
         return conn
 
@@ -1612,6 +1809,22 @@ class _TestPasswordAuth(ServerTestCase):
         with self.assertRaises(asyncssh.PermissionDenied):
             await self._connect_password('pw', 'oldpw', 'badpw', 'pw')
 
+    @asynctest
+    async def test_disabled_trivial_password_auth(self):
+        """Test disabling trivial auth with password authentication"""
+
+        async with self.connect(username='pw', password='pw',
+                                client_keys=None, disable_trivial_auth=True):
+            pass
+
+    @asynctest
+    async def test_disabled_trivial_password_change(self):
+        """Test disabling trivial aith with password change"""
+
+        async with self._connect_password('pw', 'oldpw', 'oldpw', 'pw',
+                                          disable_trivial_auth=True):
+            pass
+
 
 class _TestPasswordAsyncServerAuth(_TestPasswordAuth):
     """Unit tests for password authentication with async server callbacks"""
@@ -1632,6 +1845,15 @@ class _TestKbdintAuth(ServerTestCase):
 
         return await cls.create_server(_KbdintServer)
 
+    @asynctest
+    async def test_get_server_auth_methods(self):
+        """Test getting auth methods from the test server"""
+
+        auth_methods = await asyncssh.get_server_auth_methods(
+            self._server_addr, self._server_port, username='none')
+
+        self.assertEqual(auth_methods, ['keyboard-interactive'])
+
     @async_context_manager
     async def _connect_kbdint(self, username, responses, test_async=False):
         """Open a connection to test keyboard-interactive auth"""
@@ -1648,6 +1870,14 @@ class _TestKbdintAuth(ServerTestCase):
 
         return conn
 
+    @asynctest
+    async def test_kbdint_auth_no_prompts(self):
+        """Test keyboard-interactive authentication with no prompts"""
+
+        async with self.connect(username='none', password='kbdint',
+                                client_keys=None):
+            pass
+
     @asynctest
     async def test_kbdint_auth_password(self):
         """Test keyboard-interactive authentication via password"""
@@ -1717,6 +1947,22 @@ class _TestKbdintAuth(ServerTestCase):
         with self.assertRaises(asyncssh.PermissionDenied):
             await self._connect_kbdint('kbdint', ['badpw'])
 
+    @asynctest
+    async def test_disabled_trivial_kbdint_auth(self):
+        """Test disabled trivial auth with keyboard-interactive auth"""
+
+        async with self.connect(username='pw', password='kbdint',
+                                client_keys=None, disable_trivial_auth=True):
+            pass
+
+    @asynctest
+    async def test_disabled_trivial_kbdint_no_prompts(self):
+        """Test disabled trivial with with no keyboard-interactive prompts"""
+
+        with self.assertRaises(asyncssh.PermissionDenied):
+            await self.connect(username='none', password='kbdint',
+                               client_keys=None, disable_trivial_auth=True)
+
 
 class _TestKbdintAsyncServerAuth(_TestKbdintAuth):
     """Unit tests for keyboard-interactive auth with async server callbacks"""
